package org.nakedobjects.nof.boot;

import org.nakedobjects.noa.security.AuthenticationRequest;
import org.nakedobjects.nof.boot.system.HttpServerMonitor;
import org.nakedobjects.nof.boot.system.InstallerLookup;
import org.nakedobjects.nof.boot.system.NakedObjectsSystem;
import org.nakedobjects.nof.boot.system.SocketServerMonitor;
import org.nakedobjects.nof.core.client.ClientInstaller;
import org.nakedobjects.nof.core.conf.Configuration;
import org.nakedobjects.nof.core.conf.DefaultConfigurationLoader;
import org.nakedobjects.nof.core.context.StaticContext;
import org.nakedobjects.nof.core.context.ThreadContext;
import org.nakedobjects.nof.core.security.AuthenticatorInstaller;
import org.nakedobjects.nof.core.security.PasswordAuthenticationRequest;
import org.nakedobjects.nof.core.system.ClientConnectionInstaller;
import org.nakedobjects.nof.core.system.Installer;
import org.nakedobjects.nof.core.system.ObjectLoaderInstaller;
import org.nakedobjects.nof.core.system.ObjectPersistorInstaller;
import org.nakedobjects.nof.core.system.ServerListenerInstaller;

import java.util.Properties;

import org.apache.commons.cli.BasicParser;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.log4j.Level;


/**
 * Starts up a naked objects system with the components specified on the command line, if any. This class will
 * create an instance of NakedObjectsSystem and inject any relevant components component loaders that are need
 * in accordance with the parameters specified on the command line.
 */
public final class NakedObjects {
    private static final String HELP_OPT = "h";
    private static final String HELP_LONG_OPT = "help";
    private static final String CONFIGURATION_OPT = "c";
    private static final String CONFIGURATION_LONG_OPT = "config";
    private static final String CONNECTION_OPT = "x";
    private static final String CONNECTION_LONG_OPT = "connection";
    private static final String PERSISTOR_OPT = "r";
    private static final String PERSISTOR_LONG_OPT = "persistor";
    private static final String LOADER_OPT = "l";
    private static final String LOADER_LONG_OPT = "loader";
    private static final String VIEWER_OPT = "v";
    private static final String VIEWER_LONG_OPT = "viewer";
    private static final String TYPE_OPT = "t";
    private static final String TYPE_LONG_OPT = "type";
    private static final String NO_SPLASH_OPT = "s";
    private static final String NO_SPLASH_LONG_OPT = "nosplash";
    private static final String DIAGNOSTICS_OPT = "diagnostics";
    private static final String VERSION_OPT = "version";
    private static final String DEBUG_OPT = "debug";
    private static final String VERBOSE_OPT = "verbose";
    private static final String QUIET_OPT = "quiet";
    private static final String USER_OPT = "u";
    private static final String USER_LONG_OPT = "user";
    private static final String PASSWORD_OPT = "p";
    private static final String PASSWORD_LONG_OPT = "password";
    private static final String FIXTURE_OPT = "f";
    private static final String FIXTURE_LONG_OPT = "fixture";
    private static final String TYPE_STANDALONE = "standalone";
    private static final String TYPE_SERVER = "server";
    private static final String TYPE_CLIENT = "client";
    private static final String TYPE_PROTOTYPE = "prototype";
    private static final String TYPE_EXPLORATION = "exploration";
    private static final String ADDITIONAL_PROPERTY = "D";

    public static void main(final String[] args) {
    	new NakedObjects().run(args);
    }
    public NakedObjectsSystem run(String[] args) {

        final Options options;
        options = new Options();
        addBasicOptions(options);

        CommandLineParser parser = new BasicParser();
        CommandLine cmd = null;
        Level level = loggingLevel(args);

        final NakedObjectsSystem system = new NakedObjectsSystem(Configuration.DIRECTORY, level);
        InstallerLookup installers = system.getInstallerFactory();

        addComponentOptions(options, installers);
        try {
            cmd = parser.parse(options, args);
        } catch (ParseException e) {
            System.out.println("Error: " + e.getMessage());
            printHelp(options);
            return null;
        }

        // help option
        if (cmd.hasOption(HELP_OPT)) {
            printHelp(options);
            return null;
        }

        if (cmd.hasOption(VERSION_OPT)) {
            NakedObjectsSystem.printVersion();
            return null;
        }

        if (cmd.hasOption(DIAGNOSTICS_OPT)) {
            NakedObjectsSystem.printDiagnostics();
            return null;
        }

        String mode = cmd.getOptionValue(TYPE_OPT);
        mode = mode == null ? TYPE_PROTOTYPE : mode.toLowerCase();
        final boolean isExploration = mode.equals(TYPE_EXPLORATION);
        final boolean isPrototype = mode.equals(TYPE_PROTOTYPE);
        final boolean isClient = mode.equals(TYPE_CLIENT);
        final boolean isServer = mode.equals(TYPE_SERVER);
        final boolean isStandalone = mode.equals(TYPE_STANDALONE);

        String connection = cmd.getOptionValue(CONNECTION_OPT);
        if (!(isClient || isServer) && (connection != null && !connection.equals(""))) {
            printErrorAndHelp(options, "Error: can only specify a connection for clients and servers\n");
            return null;
        }

        final DefaultConfigurationLoader configurationLoader = configurationLoader(cmd, installers);
        if (mode != null) {
            configurationLoader.addConfigurationFile(mode + ".properties", false);
        }
        ClientConnectionInstaller clientConnectionInstaller = null;
        if (isClient || isServer) {
            system.loadConfiguration(configurationLoader);
            clientConnectionInstaller = installers.clientConnectionInstaller(connection,
                    InstallerLookup.CLIENT_CONNECTION_DEFAULT);
        }
        if (clientConnectionInstaller != null) {
            configurationLoader.addConfigurationFile(clientConnectionInstaller.getName() + ".properties", false);
            if (isClient) {
                Properties properties = clientConnectionInstaller.getRemoteProperties();
                configurationLoader.add(properties);
            }
        }

        system.setReflectorInstaller(installers.reflectorInstaller(null));
        system.setServicesInstaller(installers.servicesInstaller(null));
        system.setFixtureInstaller(installers.fixturesInstaller(cmd.getOptionValue(FIXTURE_OPT)));
        system.setObjectLoaderInstaller(installers.objectLoaderInstaller(cmd.getOptionValue(LOADER_OPT)));
        system.setTemplateImageLoader(installers.templateImageLoaderInstaller(null));

        String persistor = cmd.getOptionValue(PERSISTOR_OPT);

        if (persistor != null) {
            if (isExploration || isPrototype) {
                printErrorAndHelp(options, "Error: can't specify persistor for exploration or prototype\n");
                return null;
            } else if (isClient) {
                printErrorAndHelp(options, "Error: can't specify persistor for client\n");
                return null;
            }
        }

        ObjectPersistorInstaller persistorInstaller = installers.persistorInstaller(persistor, isServer);
        configurationLoader.addConfigurationFile(persistorInstaller.getName() + ".properties", false);
        system.setObjectPersistorInstaller(persistorInstaller);

        String viewer = cmd.getOptionValue(VIEWER_OPT);

        final ClientInstaller clientInstaller = installers.clientInstaller(viewer, isServer ? null : "dnd");
        if (clientInstaller != null) {
            configurationLoader.addConfigurationFile(clientInstaller.getName() + ".properties", false);
        }

        String[] optionValues = cmd.getOptionValues(ADDITIONAL_PROPERTY);
        if (optionValues != null) {
            for (int i = 0; i < optionValues.length; i += 2) {
                configurationLoader.add(optionValues[i], optionValues[i + 1]);
            }
        }
        system.loadConfiguration(configurationLoader);

        AuthenticatorInstaller authenticatorInstaller = installers.authenticatorInstaller(null);
        configurationLoader.addConfigurationFile(authenticatorInstaller.getName() + ".properties", false);
        system.setAuthenticatorInstaller(authenticatorInstaller);
        
        if (isClient) {
            system.setClientConnection(clientConnectionInstaller);
        }

        final String user = cmd.getOptionValue(USER_OPT);
        final String password = cmd.getOptionValue(PASSWORD_OPT);
        if (password != null && user == null) {
            printErrorAndHelp(options, "A user name must be specified with a password");
        }

        boolean noSplash = cmd.hasOption('s');
        if (noSplash) {
            system.disableSplash(true);
        }

        if (isExploration || isPrototype || isStandalone || isClient) {
            system.setContext(StaticContext.createInstance());
            system.setShareObjectsAcrossContexts(true);
        } else {
            system.setContext(ThreadContext.createInstance());
            system.setShareObjectsAcrossContexts(false);
        }
        if (isServer&& connection != null) {
            // TODO allow multiple connections by iterating through multiple connection params
            system.addListener(installers.listenerInstaller(connection));
        }

        system.setServer(isServer);
        system.init();
        system.removeSplash();

        if (clientInstaller != null) {
            runClient(system, clientInstaller, isExploration, user, password);
        }

        if (isServer) {
            runServerMonitor(system);
            runHttpServerMonitor(system);
        }
        
        return system;
    }

    private static void runClient(
            final NakedObjectsSystem system,
            final ClientInstaller clientInstaller,
            final boolean isExploration,
            final String user,
            final String password) {
        new Thread(new Runnable() {
            public void run() {
                // REVIEW is it valid to allow all types of clients to be loaded when running as a server
                // - should
                // we allow HTML but not DND
                boolean noAuthentication = isExploration;
                AuthenticationRequest authentationRequest = null;
                if (user != null) {
                    // TODO if user name set up then authenticate user here, and then set authenticateUser
                    // flag to
                    // false so not prompted. Also allow name to be passed to client to be used in its
                    // prompt
                    authentationRequest = new PasswordAuthenticationRequest(user, password);
                }
                boolean authenticateUser = !noAuthentication;
                system.runClient(clientInstaller, null, authenticateUser, isExploration, authentationRequest);
            }
        }, "Client").start();
    }

    private static void runServerMonitor(final NakedObjectsSystem system) {
        new Thread(new Runnable() {
            public void run() {
                SocketServerMonitor serverMonitor1 = new SocketServerMonitor();
                serverMonitor1.setTarget(system);
                serverMonitor1.listen();
            }
        }, "Monitor-1").start();
    }

    private static void runHttpServerMonitor(final NakedObjectsSystem system) {
        new Thread(new Runnable() {
            public void run() {
                HttpServerMonitor serverMonitor2 = new HttpServerMonitor();
                serverMonitor2.setTarget(system);
                serverMonitor2.listen();
            }
        }, "Monitor-2").start();
    }

    private static Level loggingLevel(String[] args) {
        Level level = null;
        for (int i = 0; i < args.length; i++) {
            if (args[i].equals("-" + DEBUG_OPT)) {
                level = Level.DEBUG;
                break;
            } else if (args[i].equals("-" + QUIET_OPT)) {
                level = Level.ERROR;
                break;
            } else if (args[i].equals("-" + VERBOSE_OPT)) {
                level = Level.INFO;
                break;
            }
        }
        return level;
    }

    private static DefaultConfigurationLoader configurationLoader(CommandLine cmd, InstallerLookup installers) {
        final DefaultConfigurationLoader configuration;
        configuration = new DefaultConfigurationLoader();
        installers.setConfiguration(configuration.load());

        // TODO allow for multiple properties files to be loaded
        String config = cmd.getOptionValue(CONFIGURATION_OPT);
        if (config != null) {
            configuration.addConfigurationFile(config, true);
        }
        return configuration;
    }

    private static void addBasicOptions(Options options) {
        options.addOption(HELP_OPT, HELP_LONG_OPT, false, "show this help");
        options.addOption(NO_SPLASH_OPT, NO_SPLASH_LONG_OPT, false, "don't show splash window");
        options.addOption(DEBUG_OPT, false, "print debugging messages");
        options.addOption(VERBOSE_OPT, false, "print information, warning and error messages");
        options.addOption(QUIET_OPT, false, "print error messages only");
        options.addOption(DIAGNOSTICS_OPT, false, "print information that can be used diagnose or report problems");
        options.addOption(VERSION_OPT, false, "print version information");
    }

    private static void addComponentOptions(Options options, InstallerLookup installers) {
        Option option = OptionBuilder.withArgName("config file").hasArg().withLongOpt(CONFIGURATION_LONG_OPT).withDescription(
                "read in configuration file (as well as nakedobjects.properties)").create(CONFIGURATION_OPT);
        options.addOption(option);

        option = OptionBuilder.withArgName("name").hasArg().withLongOpt(TYPE_LONG_OPT).withDescription(
                "type of system: " + TYPE_EXPLORATION + "; " + TYPE_PROTOTYPE + " (default); " + TYPE_STANDALONE + "; "
                        + TYPE_CLIENT + "; " + TYPE_SERVER).create(TYPE_OPT);
        options.addOption(option);

        option = OptionBuilder.withArgName("user name").hasArg().withLongOpt(USER_LONG_OPT).withDescription(
                "user name to log in with").create(USER_OPT);
        options.addOption(option);

        option = OptionBuilder.withArgName("password").hasArg().withLongOpt(PASSWORD_LONG_OPT).withDescription(
                "password to automatically log in with").create(PASSWORD_OPT);
        options.addOption(option);

        option = OptionBuilder.withArgName("class name").hasArg().withLongOpt(FIXTURE_LONG_OPT).withDescription(
                "fully qualified fixture class").create(FIXTURE_OPT);
        options.addOption(option);

        Object[] persistors = installers.getInstallers(ObjectPersistorInstaller.class);
        option = OptionBuilder.withArgName("name|class name").hasArg().withLongOpt(PERSISTOR_LONG_OPT).withDescription(
                "object persistor to use (ignored if type is prototype or client): " + availableComponents(persistors)
                        + "; or class name").create(PERSISTOR_OPT);
        options.addOption(option);

        Object[] objectLoaders = installers.getInstallers(ObjectLoaderInstaller.class);
        option = OptionBuilder.withArgName("name|class name").hasArg().withLongOpt(LOADER_LONG_OPT).withDescription(
                "object loader to use: " + availableComponents(objectLoaders) + "; or class name").create(LOADER_OPT);
        options.addOption(option);

        Object[] viewers = installers.getInstallers(ClientInstaller.class);
        option = OptionBuilder.withArgName("name|class name").hasArg().withLongOpt(VIEWER_LONG_OPT).withDescription(
                "viewer to use (ignored if type is server): " + availableComponents(viewers) + "; or class name").create(
                VIEWER_OPT);
        options.addOption(option);

        Object[] connections = installers.getInstallers(ServerListenerInstaller.class);
        option = OptionBuilder.withArgName("name|class name").hasArg().withLongOpt(CONNECTION_LONG_OPT).withDescription(
                "connection to use for client requests, or server to listen on: " + availableComponents(connections)).create(
                CONNECTION_OPT);
        options.addOption(option);

        Option property = OptionBuilder.withArgName("property=value").hasArg().withValueSeparator().withDescription(
                "use value for given property").create(ADDITIONAL_PROPERTY);
        property.setArgs(Option.UNLIMITED_VALUES);
        options.addOption(property);
    }

    private static StringBuffer availableComponents(Object[] factories) {
        StringBuffer types = new StringBuffer();
        for (int i = 0; i < factories.length; i++) {
            if (i > 0) {
                types.append("; ");
            }
            types.append(((Installer) factories[i]).getName());
        }
        return types;
    }

    private static void printErrorAndHelp(Options options, String error) {
        System.out.println(error);
        printHelp(options);
    }

    private static void printHelp(Options options) {
        HelpFormatter help = new HelpFormatter();
        help.defaultWidth = 80;
        help.printHelp("NakedObjects [options]", options, false);
    }
}
// Copyright (c) Naked Objects Group Ltd.
