package org.nakedobjects.nof.boot.system;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.nakedobjects.noa.NakedObjectRuntimeException;
import org.nakedobjects.nof.core.client.ClientInstaller;
import org.nakedobjects.nof.core.conf.Configuration;
import org.nakedobjects.nof.core.security.AuthenticatorInstaller;
import org.nakedobjects.nof.core.system.ClientConnectionInstaller;
import org.nakedobjects.nof.core.system.FixturesInstaller;
import org.nakedobjects.nof.core.system.Installer;
import org.nakedobjects.nof.core.system.InstanceCreationClassException;
import org.nakedobjects.nof.core.system.InstanceCreationException;
import org.nakedobjects.nof.core.system.InstanceFactory;
import org.nakedobjects.nof.core.system.ObjectLoaderInstaller;
import org.nakedobjects.nof.core.system.ObjectPersistorInstaller;
import org.nakedobjects.nof.core.system.ReflectorInstaller;
import org.nakedobjects.nof.core.system.ServerListenerInstaller;
import org.nakedobjects.nof.core.system.ServicesInstaller;
import org.nakedobjects.nof.core.system.TemplateImageLoaderInstaller;
import org.nakedobjects.nof.core.system.UnavailableClassException;
import org.nakedobjects.nof.core.util.NakedObjectConfiguration;


/**
 * This class retrieves named installers from those loaded at creation. A list of possible classes (of the
 * type {@link Installer}) are read in from the resource file framework.properties and instantiated. Each
 * installer has a name that will be compared when one of this classes methods are called.
 */
public class InstallerLookup {
    public static final String CLIENT_CONNECTION_DEFAULT = "xstream";
    public static final String PROPERTY_BASE = Configuration.ROOT + "component.";
    private static final Logger LOG = Logger.getLogger(InstallerLookup.class);
    private static final String PERSISTOR_IN_MEMORY = "in-memory";
    private static final String PERSISTOR_IN_FILES = "xml";
    private String COMPONENT_FILE = "framework.properties";
    private Vector componentInstallers;

    private NakedObjectConfiguration configuration;

    public InstallerLookup() {
        loadInstallers();
    }

    public AuthenticatorInstaller authenticatorInstaller(String name) {
        if (name == null) {
            name = configuration.getString(PROPERTY_BASE + "authenticator", "file");
        }
        return (AuthenticatorInstaller) getInstaller(name, AuthenticatorInstaller.class);
    }

    public ClientConnectionInstaller clientConnectionInstaller(final String name, final String defaultName) {
        String installerName;
        if (name == null) {
            installerName = configuration.getString(PROPERTY_BASE + "connection", defaultName);
        } else {
            installerName = name;
        }
        if (installerName == null) {
            return null;
        }
        return (ClientConnectionInstaller) getInstaller(installerName, ClientConnectionInstaller.class);
    }

    public FixturesInstaller fixturesInstaller(String name) {
        if (name == null) {
            name = configuration.getString(PROPERTY_BASE + "fixtures", "configuration-fixtures");
        }
        return (FixturesInstaller) getInstaller(name, FixturesInstaller.class);
    }

    public ServerListenerInstaller listenerInstaller(String name) {
        ServerListenerInstaller component = (ServerListenerInstaller) getInstaller(name, ServerListenerInstaller.class);
        if (name == null) {
            throw new NakedObjectRuntimeException("No server listener specified");
        }
        if (component == null) {
            throw new NakedObjectRuntimeException("No server listener of type " + name);
        }
        return component;
    }

    public ObjectLoaderInstaller objectLoaderInstaller(String name) {
        if (name == null) {
            name = configuration.getString(PROPERTY_BASE + "loader");
        }
        if (name == null) {
            return null;
        }
        return (ObjectLoaderInstaller) getInstaller(name, ObjectLoaderInstaller.class);
    }

    public ObjectPersistorInstaller persistorInstaller(final String name, final boolean isForServer) {
        final String persistorName;
        if (name == null) {
            persistorName = configuration.getString(PROPERTY_BASE + "persistor", isForServer ? PERSISTOR_IN_FILES
                    : PERSISTOR_IN_MEMORY);
        } else {
            persistorName = name;
        }

        ObjectPersistorInstaller installPersistor = (ObjectPersistorInstaller) getInstaller(persistorName,
                ObjectPersistorInstaller.class);
        return installPersistor;
    }

    public ReflectorInstaller reflectorInstaller(String name) {
        if (name == null) {
            name = configuration.getString(PROPERTY_BASE + "reflector", "java");
        }
        ReflectorInstaller installer = (ReflectorInstaller) getInstaller(name, ReflectorInstaller.class);
        return installer;
    }

    public ServicesInstaller servicesInstaller(String name) {
        ServicesInstaller installServices = null;
        if (name == null) {
            name = configuration.getString(PROPERTY_BASE + "services", "services-properties");
        }
        installServices = (ServicesInstaller) getInstaller(name, ServicesInstaller.class);
        return installServices;
    }

    public TemplateImageLoaderInstaller templateImageLoaderInstaller(String name) {
        if (name == null) {
            name = configuration.getString(PROPERTY_BASE + "image-loader", "awt-template-image-loader");
        }
        return (TemplateImageLoaderInstaller) getInstaller(name, TemplateImageLoaderInstaller.class);
    }

    public ClientInstaller clientInstaller(final String name, final String defaultName) {
        String viewer;
        if (name == null) {
            viewer = configuration.getString(PROPERTY_BASE + "viewer", defaultName);
        } else {
            viewer = name;
        }
        if (viewer == null) {
            return null;
        } else {
            return (ClientInstaller) getInstaller(viewer, ClientInstaller.class);
        }
    }

    Installer getInstaller(String name, Class cls) {
        if (name == null) {
            throw new NullPointerException("No name specified");
        }
        for (Enumeration e = componentInstallers.elements(); e.hasMoreElements();) {
            Installer comp = (Installer) e.nextElement();
            if (cls.isAssignableFrom(comp.getClass()) && comp.getName().equals(name)) {
                return comp;
            }
        }
        try {
            return (Installer) InstanceFactory.createInstance(name);
        } catch (InstanceCreationException e) {
            throw new InstanceCreationException("Specification error in " + COMPONENT_FILE, e);
        }
    }

    public Installer[] getInstallers(Class cls) {
        List list = new ArrayList();
        for (Enumeration e = componentInstallers.elements(); e.hasMoreElements();) {
            Installer comp = (Installer) e.nextElement();
            if (cls.isAssignableFrom(comp.getClass())) {
                list.add(comp);
            }
        }
        return (Installer[]) list.toArray(new Installer[list.size()]);
    }

    private void loadInstallers() {
        componentInstallers = new Vector();
        InputStream in = FrameworkPropertiesLoader.getFrameworkPropertiesStream(this, COMPONENT_FILE);
        BufferedReader reader = new BufferedReader(new InputStreamReader(in));
        try {
            String line;
            while ((line = reader.readLine()) != null) {
                String className = line.trim();
                if (className.length() == 0 || className.startsWith("#")) {
                    continue;
                }
                try {
                    Installer object = (Installer) InstanceFactory.createInstance(className);
                    LOG.debug("created component instance: " + object.getName() + " - " + className);
                    componentInstallers.addElement(object);
                } catch (UnavailableClassException e) {
                    LOG.info("component will not be available: " + className);
                } catch (InstanceCreationClassException e) {
                    LOG.info("instance creation exception: " + e.getMessage());
                } catch (InstanceCreationException e) {
                    throw e;
                    /*
                     * LOG.info("no component class found: " + className); } catch (NoClassDefFoundError e) {
                     * LOG.info("component found " + className + ", but is missing a dependent class", e); }
                     * catch (InstantiationException e) { throw new NakedObjectRuntimeException(className + "
                     * cannot be instantiated", e); } catch (IllegalAccessException e) { throw new
                     * NakedObjectRuntimeException(className + " cannot be accessed", e);
                     */
                }
            }
        } catch (IOException e) {
            throw new NakedObjectRuntimeException(e);
        } finally {
            if (reader != null) {
                try {
                    reader.close();
                } catch (IOException e) {
                    throw new NakedObjectRuntimeException(e);
                }
            }
        }
    }

    public void setConfiguration(NakedObjectConfiguration configuration) {
        this.configuration = configuration;
    }

}

// Copyright (c) Naked Objects Group Ltd.
